import re

__author__  = 'Julien Bigot <julien.bigot@cea.fr>'
__version__ = '0.1-wip'
__date__    = '2013/01/01'
__license__ = 'CECILL-B'

class ParsingError(StandardError):
    def __init__(self, data, index, expected=None):
        self._data = data
        self._index = index
        self._expected = set()
        if expected is not None:
            self._expected.add(expected)

    def __str__(self):
        result = 'unexpected data in data file, line '+str(self._index[1])+', column '+str(self._index[2])+' '
        if len(self._expected) > 0:
            result += 'expected '+' or '.join(self._expected)+', '
        result += 'found:\n> '+(self._data[self._index[0]:self._index[0]+50]).replace('\n','\n> ')+'...'
        return result

class BasicParser:
    """ A simple ll parser based on plain python functions """

    def __init__(self, data=None):
        if ( not data is None ):
            self.setData(data)

    def setData(self, data):
        """
        Sets the content of the data file.

        data can be either a string containing the data or a file object
        """
        if (isinstance(data, basestring)):
            self._rawData = data
        else:
            self._rawData = data.read()
        self._index = None
        self._var = None
        self._fort = None
        self._index = ( 0, 1, 1 )
        self._error = None
        err = self._parse()
        if self._index[0] != len(self._rawData):
            if self._err(err):
                raise err
            self._invalid('')
    
    def _err(self, obj):
        return isinstance(obj, ParsingError)
    
    def _or(self, *errs):
        error = None
        for obj in errs:
            if not self._err(obj): continue
            if error is None: error = obj
            elif error._index[0] < obj._index[0]: error = obj
            elif error._index[0] == obj._index[0]: error._expected |= obj._expected
        return error
    
    def _required(self, rewindIndex, err):
        self._index = rewindIndex
        return err

    def _invalid(self, expected):
        return ParsingError(self._rawData, self._index, expected)

    def _parse(self):
        raise NotImplementedError

    def _token(self, regexp, expected):
        regex = re.compile(regexp, re.M|re.S)
        res = regex.match(self._rawData, self._index[0])
        if res:
            index = res.end()
            res = self._rawData[res.start():res.end()]
            nlCount = res.count('\n')
            line = self._index[1] + nlCount
            if ( nlCount > 0 ): lineIndex = len(res) - res.rfind('\n')
            else: lineIndex = self._index[2] + len(res)
            self._index = ( index, line, lineIndex )
            return res
        return self._invalid(expected)
