#!/bin/bash

#     Usage: deadcode.sh [-s] [<file>...]
#        -s: sort unused procedures by file
# <file>...: list of files of the program, defaults to all *.[fF]90 in src
# WARNING: there might be some false positive for example if the procedure is used from C

cd "$(dirname "$0")/.."

SORT="0"
if [ "$1" = "-s" ]
then
	SORT="1"
	shift
fi

SRCDIR="src"
if [ $# -gt 0 ]; then
	SRCDIR="$1"
fi

ALLFILES="$(find "${SRCDIR}" -iname '*.f90')"

GYSELA_CODE="$(mktemp --tmpdir)"
DEFINED_PROCEDURES="$(mktemp --tmpdir)"
CALLED_PROCEDURES="$(mktemp --tmpdir)"
INTERFACED_PROCEDURES="$(mktemp --tmpdir)"
USED_PROCEDURES="$(mktemp --tmpdir)"
UNUSED_PROCEDURES_S0="$(mktemp --tmpdir)"
UNUSED_PROCEDURES_S1="$(mktemp --tmpdir)"

sed ':a;N;$!ba;s/\&\s*\(![^\n]*\)\?\n/ /g' $ALLFILES | tr '[:upper:]' '[:lower:]' > "${GYSELA_CODE}"
egrep '^\s*(subroutine|function|interface)\b' "${GYSELA_CODE}" | awk '{print $2}' | sed 's/(.*//' | sort -u > "${DEFINED_PROCEDURES}"
grep '\bcall\b' "${GYSELA_CODE}" | sed 's/^.*\bcall\b\s*\b\([a-zA-Z0-9_]*\)\b.*$/\1/' > "${CALLED_PROCEDURES}"
grep '\bmodule\s\+procedure\b' "${GYSELA_CODE}" | sed 's/.*\bmodule\s\+procedure\b\(.*\)\(!.*\)\?/\1/' | tr ',' '\n' | sed 's/^\s*\b\([a-zA-Z0-9_]*\)\s*$/\1/' > "${INTERFACED_PROCEDURES}"
grep -v -e '^#' 'doc/tokeep.func' -e '^\s*$' | gawk '{print $1}' >> "${CALLED_PROCEDURES}"
sort -u "${CALLED_PROCEDURES}" "${INTERFACED_PROCEDURES}" > "${USED_PROCEDURES}"
diff -i '--old-line-format=%L' '--new-line-format=' '--unchanged-line-format=' "${DEFINED_PROCEDURES}" "${USED_PROCEDURES}" > "${UNUSED_PROCEDURES_S0}"
cat "${UNUSED_PROCEDURES_S0}" | while read UNUSED_SYMBOL; do
	if fgrep "${UNUSED_SYMBOL}" "${GYSELA_CODE}" 2>/dev/null | egrep -v '^\s*(subroutine|function)\b' | grep "\b${UNUSED_SYMBOL}\s*(" &>/dev/null
	then true
	else echo "${UNUSED_SYMBOL}" >> "${UNUSED_PROCEDURES_S1}"
	fi
done 
if [ "${SORT}" -eq "1" ]
then
	cat "${UNUSED_PROCEDURES_S1}" | while read UNUSED_SYMBOL; do
		egrep -i "\b${UNUSED_SYMBOL}\b" ${ALLFILES} | gawk "{sub(/:.*/,\"\"); print \$1\" ${UNUSED_SYMBOL}\"}" | sort -u
	done | sort -u | gawk '{ print $2" :\t "$1}'
else
	cat "${UNUSED_PROCEDURES_S1}"
fi

rm -f "${GYSELA_CODE}" "${DEFINED_PROCEDURES}" "${CALLED_PROCEDURES}" "${INTERFACED_PROCEDURES}" "${USED_PROCEDURES}" "${UNUSED_PROCEDURES_S0}" "${UNUSED_PROCEDURES_S1}"