#!/bin/bash
#**************************************************************
#  Copyright Euratom-CEA
#  Authors : 
#     Virginie Grandgirard (virginie.grandgirard@cea.fr)
#     Chantal Passeron (chantal.passeron@cea.fr)
#     Guillaume Latu (guillaume.latu@cea.fr)
#     Xavier Garbet (xavier.garbet@cea.fr)
#     Philippe Ghendrih (philippe.ghendrih@cea.fr)
#     Yanick Sarazin (yanick.sarazin@cea.fr)
#  
#  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
#  is a 5D gyrokinetic global full-f code for simulating 
#  the plasma turbulence in a tokamak.
#  
#  This software is governed by the CeCILL-B license 
#  under French law and abiding by the rules of distribution 
#  of free software.  You can  use, modify and redistribute 
#  the software under the terms of the CeCILL-B license as 
#  circulated by CEA, CNRS and INRIA at the following URL
#  "http://www.cecill.info". 
#**************************************************************
((k=0))

for fich in $*; do
    if [ ! -f "$fich" ]; then
        echo "Fichier inexistant : " $fich
        exit 1
    fi

    depend=$(tr '[:upper:]' '[:lower:]' < $fich | sed -n "s/^\s*use  *\([a-z0-9_]*\).*$/\1/p"  | sort -u  | tr "\n" " ")

    includ=$(grep -i -E '^(#|\s*)include' $fich | sed -e "s%^.*include[^\"']*[\"']\([^\"']*\)[\"'].*$%\1%" | grep -v '^mpif\.h$' | grep -v '^mfdiag_inc\.h$' |sort -u)

#    printf "depend $fich $depend"
#    printf "%s, include = %s %s %s %s %s\n" $fich $includ

    mod=$(tr '[:upper:]' '[:lower:]' < $fich | awk '/^module/ {print $2}')
    if [ ${#mod} -eq 0 ]; then
        mod=${fich%.*90}
    fi
#    printf "mod =$mod=\n"
    for smo in $mod; do
        depmod[k]=${depend%' '}
        fichmod[k]=$smo
        fichlist[k]=${fich%.*90}
        sufflist[k]=.${fich#*.}
        if [ -n "$includ" ]; then
        finclud[k]=$includ
        fi
        ((k++))
    done
done

#((i=0))
#while [ $i -lt $k ]; do
#  printf "Liste =%s / %s= %s %s\n" "${depmod[i]}" "${fichmod[i]}" "${fichlist[i]}" "${sufflist[i]}"
#  ((i++))
#done
#printf "k =$k=\n"
((i=1))
while [ $i -le $k ]; do
    prefich=${fichlist[$i-1]}
    sufx=${sufflist[$i-1]}
    deplist=(${depmod[$i-1]})
    nn=${#deplist[*]}
    ((j=1))
    while [ $j -le $nn ]; do
        ((target=-1))
        ((q=1))
        while [ $q -le $k ]; do
           if [ "${deplist[j-1]}" = "${fichmod[q-1]}" ]; then
               ((target=q-1))
           fi
           ((q++))
#            printf "$i $q $prefich ${deplist[j-1]} = ${fichmod[q-1]}\n"
        done
        if [ $target -ne -1 ]; then
            intdeplist[$j-1]=$target
        else
            intdeplist[$j-1]="-1"
#             echo "unable to find module ${deplist[j-1]} used in file $prefich${sufx}" 1>&2
        fi
        ((j++))
    done
    printf "%20s.o : %s${sufx}" "$prefich" "$prefich"
    ((m=1))
    ((j=1))
    while [ $j -le $nn ]; do
        if [ $m -gt 3 ]; then
            printf "\\" 
            printf "\n%24s" " " 
            ((m=0))
        fi
        if [ ${intdeplist[j-1]} -ge 0 ]
        then
            printf " %s.o" ${fichlist[${intdeplist[j-1]}]}
        fi
        ((j++))
        ((m++))
    done
    ((m++))
    if [ -n "${finclud[$i-1]}" ]; then
        printf "\\" 
        printf "\n%24s" " "
        printf " %s " ${finclud[$i-1]}
#        printf " ${finclud[$i-1]}"
    fi
    printf "\n"
    ((i++))
done
