#include "config-c.h"

#include <assert.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <execinfo.h>
#include <stdlib.h>
#include <unistd.h>

#define SIZE 100

static const uint64_t FNV0_PRIME_8 = ((uint64_t)1<<40) + ((uint64_t)1<<8) + (0xb3);


/** Computes and dumps the FNV-0 hash of an array
 * \param data the data to hash
 * \param size the size of one element in data
 * \param count the number of elements in data
 * \param result the hash as a 30 char string
 * \param result_len length of result, should always be 30
 */
void FORTSYM(fnv0_impl, FNV0_IMPL)(const void* data, const int* size, const int* count, char* result, int result_len)
{
	assert(result_len == 30);
	int ii;
	
	uint64_t fnv0 = 0;
	for ( ii=0; ii<(*count)*(*size); ++ii ) {
		fnv0 *= FNV0_PRIME_8;
		fnv0 ^= ((uint8_t*)data)[ii];
	}
	
	char hexfnv0[17];
	for ( ii=15; ii>=0; --ii ) {
		if ( (fnv0&0xF) < 10 ) hexfnv0[ii] = '0'+(fnv0&0xF);
		else hexfnv0[ii] = 'A'-10+(fnv0&0xF);
		fnv0 >>= 4;
	}
	hexfnv0[16] = 0;
	
	memset(result, ' ', 30);
	ii = snprintf(result, 30, "0x%s:%-10d", hexfnv0, *count);
	assert(ii<=30);
	result[29] = ' ';
}

// void FORTSYM(backtrace, BACKTRACE)(int fd) {
// 
//   int j, nptrs;
//   void *buf[SIZE];
//   char **strings;
//   
//   nptrs = backtrace(buf, SIZE);
//   /* printf("backtrace() returned %d addresses\n", nptrs); */
// 
//   strings = backtrace_symbols(buf, nptrs);
//   if (strings == NULL) {
//     perror("backtrace_symbols");
//   }
// 
//   for (j = 0; j < nptrs; j++)
//     write(fd, strings[j], strlen(strings[j]));
// 
//   free(strings);
// 
//   return;
// 
// }
