!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!-------------------------------------------------------
! file : OMPutils.F90
! date : 20/03/2003
!  subroutines used for the OpenMP parallelization
!---------------------------------------------
module OMPutils_module
  use prec_const
  use wrap_alloc_module

  implicit none

#ifdef _OPENMP
  integer, public :: omp_get_thread_num, omp_get_num_threads
  external omp_get_thread_num, omp_get_num_threads
#endif

  type, public :: omp1dvector
     real(RKIND), dimension(:), pointer :: val
  end type omp1dvector
  type, public :: omp1cvector
     complex(CKIND), dimension(:), pointer :: val
  end type omp1cvector
  type, public :: omp2dvector
     real(RKIND), dimension(:,:), pointer :: val
  end type omp2dvector
  type, public :: omp2cvector
     complex(CKIND), dimension(:,:), pointer :: val
  end type omp2cvector
  type, public :: omp2ivector
     integer, dimension(:,:), pointer :: val
  end type omp2ivector
  type, public :: omp3dvector
     real(RKIND), dimension(:,:,:), pointer :: val
  end type omp3dvector
  type, public :: omp4dvector
     real(RKIND), dimension(:,:,:,:), pointer :: val
  end type omp4dvector

  public
  !-> real 1D arrays
  !--> radial direction
  type(omp1dvector), dimension (:), pointer, public :: Romp1_0Nr
  type(omp1dvector), dimension (:), pointer, public :: Romp2_0Nr
  type(omp1dvector), dimension (:), pointer, public :: Romp3_0Nr
  type(omp1dvector), dimension (:), pointer, public :: Romp4_0Nr
  type(omp1dvector), dimension (:), pointer, public :: Romp5_0Nr
  type(omp1dvector), dimension (:), pointer, public :: Romp6_0Nr
  type(omp1dvector), dimension (:), pointer, public :: Romp7_0Nr
  type(omp1dvector), dimension (:), pointer, public :: Romp8_0Nr
  type(omp1dvector), dimension (:), pointer, public :: Romp1_1Nrp1
  type(omp1dvector), dimension (:), pointer, public :: Romp2_1Nrp1
  type(omp1dvector), dimension (:), pointer, public :: Romp3_1Nrp1
  type(omp1dvector), dimension (:), pointer, public :: Romp4_1Nrp1
  !--> theta direction 
  type(omp1dvector), dimension (:), pointer, public :: Romp1_0Ntheta
  type(omp1dvector), dimension (:), pointer, public :: Romp2_0Ntheta
  type(omp1dvector), dimension (:), pointer, public :: Romp3_0Ntheta
  type(omp1dvector), dimension (:), pointer, public :: Romp4_0Ntheta
  type(omp1dvector), dimension (:), pointer, public :: Romp5_0Ntheta
  type(omp1dvector), dimension (:), pointer, public :: Romp6_0Ntheta
  type(omp1dvector), dimension (:), pointer, public :: &
    Romp1_1Nthetap1
  type(omp1dvector), dimension (:), pointer, public :: &
    Romp2_1Nthetap1
  !--> phi direction 
  type(omp1dvector), dimension (:), pointer, public :: Romp1_0Nphi
  type(omp1dvector), dimension (:), pointer, public :: Romp2_0Nphi
  type(omp1dvector), dimension (:), pointer, public :: Romp3_0Nphi
  type(omp1dvector), dimension (:), pointer, public :: Romp4_0Nphi
  type(omp1dvector), dimension (:), pointer, public :: Romp5_0Nphi
  type(omp1dvector), dimension (:), pointer, public :: Romp6_0Nphi
  !--> vparallel direction
  type(omp1dvector), dimension (:), pointer, public :: Romp1_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp2_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp3_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp4_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp5_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp6_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp7_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp8_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp9_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp10_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp11_0Nvpar
  type(omp1dvector), dimension (:), pointer, public :: Romp12_0Nvpar
  !--> used for 2D advections
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp_vect_mpstencil
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp_fillde_112
  !--> used for cubic splines
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp_deriv_01
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp_scoefr
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp_scoeftheta
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp_scoefphi
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp_scoefvpar
  !-> used for PSM scheme
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp_alpha
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp_rhs
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp_vect
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp1_0Nrp1
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp2_0Nrp1
  type(omp1dvector), dimension (:), pointer, &
    public :: Romp1_0Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp1_0Nrp1_0Ntheta
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp2_0Nrp1_0Ntheta
  !-> complex 1D arrays
  type(omp1cvector), dimension (:), pointer, &
    public :: Comp_0Nr

  !-> real 2D arrays 
  !----> (Nr,Ntheta)
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp1_0Nr_0Ntheta
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp2_0Nr_0Ntheta
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp1_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp2_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp3_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp4_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp5_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp6_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp9_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp10_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp11_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp12_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp13_1Nrp1_1Nthetap1
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp14_1Nrp1_1Nthetap1
  !----> (1:Nr+buf,1:Ntheta+buf)
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp1_1Nrpb_1Nthetapb
  !----> (1:nbmoments,Nr)
  type(omp2dvector), dimension (:), pointer, &
    public :: Romp1_1nbmom_0Nr

  !-> complex 2D arrays 
  !----> (1:Nr+1,1:Ntheta+1)
  type(omp2cvector), dimension (:), pointer, &
    public :: Comp1_1Nrp1_1Ntheta
  type(omp2cvector), dimension (:), pointer, &
    public :: Comp2_1Nrp1_1Ntheta
  !----> (1:Ntheta,1:Nr+1)
  type(omp2cvector), dimension (:), pointer, &
    public :: Comp1_1Ntheta_1Nrp1
  !----> (1:Ntheta+buf,1:Nr+buf)
  type(omp2cvector), dimension (:), pointer, &
    public :: Comp1_1Nthetapb_1Nrpb
  !----> (1:Nr+buf,1:Ntheta+buf)
  type(omp2cvector), dimension (:), pointer, &
    public :: Comp1_1Nrpb_1Nthetapb

  integer, private :: max_dim

  !******************************
  contains
  !******************************   

  !-----------------------------------------------------
  ! Constructor
  !-----------------------------------------------------
  subroutine init_OMP
    use globals, only : Nr, Ntheta, Nphi, Nvpar, Nmu, &
      Nbthread, stencil, nbmoments_GC, nbmoments_part, &
      hffilterfreq
    use globals, only : conservative_scheme
!R3 #include "code_instrumentation/r3_info.h" !R3

    integer :: tid
    integer :: nbmoments

!R3 call r3_info_begin (r3_info_index_1, 'Init_OMP') !R3
    call Write_dyn_balise_open('Init_OMP') !R4

    call write_param("nbmoments_GC", nbmoments_GC)
    call write_param("nbmoments_part", nbmoments_part)
    
    nbmoments = max(nbmoments_GC,nbmoments_part)
    call write_expr("nbmoments", "max(nbmoments_GC,nbmoments_part)" )
    !*** structure allocations ***
    !-> allocation of structures for real 1D arrays
    !--> radial direction
    allocate(Romp1_0Nr(1:Nbthread))
    allocate(Romp2_0Nr(1:Nbthread))
    allocate(Romp3_0Nr(1:Nbthread))
    allocate(Romp4_0Nr(1:Nbthread))
    allocate(Romp5_0Nr(1:Nbthread))
    allocate(Romp6_0Nr(1:Nbthread))
    allocate(Romp7_0Nr(1:Nbthread))
    allocate(Romp8_0Nr(1:Nbthread))
    allocate(Romp1_1Nrp1(1:Nbthread))
    allocate(Romp2_1Nrp1(1:Nbthread))
    allocate(Romp3_1Nrp1(1:Nbthread))
    allocate(Romp4_1Nrp1(1:Nbthread))
    !-> theta direction
    allocate(Romp1_0Ntheta(1:Nbthread))
    allocate(Romp2_0Ntheta(1:Nbthread))
    allocate(Romp3_0Ntheta(1:Nbthread))
    allocate(Romp4_0Ntheta(1:Nbthread))
    allocate(Romp5_0Ntheta(1:Nbthread))
    allocate(Romp6_0Ntheta(1:Nbthread))
    allocate(Romp1_1Nthetap1(1:Nbthread))
    allocate(Romp2_1Nthetap1(1:Nbthread))
    !-> phi direction
    allocate(Romp1_0Nphi(1:Nbthread))
    allocate(Romp2_0Nphi(1:Nbthread))
    allocate(Romp3_0Nphi(1:Nbthread))
    allocate(Romp4_0Nphi(1:Nbthread))
    allocate(Romp5_0Nphi(1:Nbthread))
    allocate(Romp6_0Nphi(1:Nbthread))
    !-> vparallel direction
    allocate(Romp1_0Nvpar(1:Nbthread))
    allocate(Romp2_0Nvpar(1:Nbthread))
    allocate(Romp3_0Nvpar(1:Nbthread))
    allocate(Romp4_0Nvpar(1:Nbthread))
    allocate(Romp5_0Nvpar(1:Nbthread))
    allocate(Romp6_0Nvpar(1:Nbthread))
    allocate(Romp7_0Nvpar(1:Nbthread))
    allocate(Romp8_0Nvpar(1:Nbthread))
    allocate(Romp9_0Nvpar(1:Nbthread))
    allocate(Romp10_0Nvpar(1:Nbthread))
    allocate(Romp11_0Nvpar(1:Nbthread))
    allocate(Romp12_0Nvpar(1:Nbthread))
    !-> used for 2D advections
    allocate(Romp_vect_mpstencil(1:Nbthread))
    allocate(Romp_fillde_112(1:Nbthread))
    !--> used for cubic splines
    allocate(Romp_deriv_01(1:Nbthread))
    allocate(Romp_scoefr(1:Nbthread))
    allocate(Romp_scoeftheta(1:Nbthread))
    allocate(Romp_scoefphi(1:Nbthread))
    allocate(Romp_scoefvpar(1:Nbthread))
    !--> used for PSM scheme
    if (conservative_scheme) then
      allocate(Romp_alpha(1:Nbthread))
      allocate(Romp_rhs(1:Nbthread))
      allocate(Romp_vect(1:Nbthread))
      allocate(Romp1_0Nrp1(1:Nbthread))
      allocate(Romp2_0Nrp1(1:Nbthread))
      allocate(Romp1_0Nthetap1(1:Nbthread))
      allocate(Romp1_0Nrp1_0Ntheta(1:Nbthread))
      allocate(Romp2_0Nrp1_0Ntheta(1:Nbthread))
    end if
    !-> allocation of structures for complex 1D arrays
    allocate(Comp_0Nr(1:Nbthread))

    call write_param("stencil", stencil)

    allocate(Romp1_1Nrp1_1Nthetap1(1:Nbthread))
    allocate(Romp2_1Nrp1_1Nthetap1(1:Nbthread))
    allocate(Romp3_1Nrp1_1Nthetap1(1:Nbthread))
    allocate(Romp4_1Nrp1_1Nthetap1(1:Nbthread))
    allocate(Romp5_1Nrp1_1Nthetap1(1:Nbthread))
    allocate(Romp6_1Nrp1_1Nthetap1(1:Nbthread))

    allocate(Comp1_1Nrp1_1Ntheta(1:Nbthread))
    allocate(Comp2_1Nrp1_1Ntheta(1:Nbthread))

    allocate(Romp9_1Nrp1_1Nthetap1(0:MAX(Nmu,Nbthread)))
    allocate(Romp10_1Nrp1_1Nthetap1(0:MAX(Nmu,Nbthread)))
    allocate(Romp11_1Nrp1_1Nthetap1(0:MAX(Nmu,Nbthread)))
    allocate(Romp12_1Nrp1_1Nthetap1(0:MAX(Nmu,Nbthread)))
    allocate(Romp13_1Nrp1_1Nthetap1(0:MAX(Nmu,Nbthread)))
    allocate(Romp14_1Nrp1_1Nthetap1(0:MAX(Nmu,Nbthread)))

    allocate(Romp1_0Nr_0Ntheta(1:Nbthread))
    allocate(Romp2_0Nr_0Ntheta(1:Nbthread))

    !-> allocation of structures for real 2D arrays (Nr,Ntheta)
    if (hffilterfreq .ne. 0) then
      allocate(Romp1_1Nrpb_1Nthetapb(1:Nbthread))
    end if
    allocate(Romp1_1nbmom_0Nr(1:Nbthread))
    !-> allocation of structures for complex 2D arrays (Nr,Ntheta)
    if (hffilterfreq .ne. 0) then
      allocate(Comp1_1Nrpb_1Nthetapb(1:Nbthread))
      allocate(Comp1_1Nthetapb_1Nrpb(1:Nbthread))
    end if
    allocate(Comp1_1Ntheta_1Nrp1(1:Nbthread))

    !*** array allocations ***
    !-> real 1D arrays
    do tid = 1,Nbthread
      !--> radial direction
      call FTR_ALLOC_1_ZERO(Romp1_0Nr(tid)%val,0,Nr)
      call FTR_ALLOC_1_ZERO(Romp2_0Nr(tid)%val,0,Nr)
      call FTR_ALLOC_1_ZERO(Romp3_0Nr(tid)%val,0,Nr)
      call FTR_ALLOC_1_ZERO(Romp4_0Nr(tid)%val,0,Nr)
      call FTR_ALLOC_1_ZERO(Romp5_0Nr(tid)%val,0,Nr)
      call FTR_ALLOC_1_ZERO(Romp6_0Nr(tid)%val,0,Nr)
      call FTR_ALLOC_1_ZERO(Romp7_0Nr(tid)%val,0,Nr)
      call FTR_ALLOC_1_ZERO(Romp8_0Nr(tid)%val,0,Nr)
      call FTR_ALLOC_1_ZERO(Romp1_1Nrp1(tid)%val,1,Nr+1)
      call FTR_ALLOC_1_ZERO(Romp2_1Nrp1(tid)%val,1,Nr+1)
      call FTR_ALLOC_1_ZERO(Romp3_1Nrp1(tid)%val,1,Nr+1)
      call FTR_ALLOC_1_ZERO(Romp4_1Nrp1(tid)%val,1,Nr+1)
      !--> theta direction 
      call FTR_ALLOC_1_ZERO(Romp1_0Ntheta(tid)%val,0,Ntheta)
      call FTR_ALLOC_1_ZERO(Romp2_0Ntheta(tid)%val,0,Ntheta)
      call FTR_ALLOC_1_ZERO(Romp3_0Ntheta(tid)%val,0,Ntheta)
      call FTR_ALLOC_1_ZERO(Romp4_0Ntheta(tid)%val,0,Ntheta)
      call FTR_ALLOC_1_ZERO(Romp5_0Ntheta(tid)%val,0,Ntheta)
      call FTR_ALLOC_1_ZERO(Romp6_0Ntheta(tid)%val,0,Ntheta)
      call FTR_ALLOC_1_ZERO(Romp1_1Nthetap1(tid)%val,1,Ntheta+1)
      call FTR_ALLOC_1_ZERO(Romp2_1Nthetap1(tid)%val,1,Ntheta+1)
      !--> phi direction 
      call FTR_ALLOC_1_ZERO(Romp1_0Nphi(tid)%val,0,Nphi)
      call FTR_ALLOC_1_ZERO(Romp2_0Nphi(tid)%val,0,Nphi)
      call FTR_ALLOC_1_ZERO(Romp3_0Nphi(tid)%val,0,Nphi)
      call FTR_ALLOC_1_ZERO(Romp4_0Nphi(tid)%val,0,Nphi)
      call FTR_ALLOC_1_ZERO(Romp5_0Nphi(tid)%val,0,Nphi)
      call FTR_ALLOC_1_ZERO(Romp6_0Nphi(tid)%val,0,Nphi)
      !--> vparallel direction 
      call FTR_ALLOC_1_ZERO(Romp1_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp2_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp3_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp4_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp5_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp6_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp7_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp8_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp9_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp10_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp11_0Nvpar(tid)%val,0,Nvpar)
      call FTR_ALLOC_1_ZERO(Romp12_0Nvpar(tid)%val,0,Nvpar)
      !--> used for 2D advections
      call FTR_ALLOC_1_ZERO(Romp_fillde_112(tid)%val,1,12)
      call FTR_ALLOC_1_ZERO(Romp_vect_mpstencil(tid)%val,-stencil,stencil)
      !--> used for cubic splines
      call FTR_ALLOC_1_ZERO(Romp_deriv_01(tid)%val,0,1)
      call FTR_ALLOC_1_ZERO(Romp_scoefr(tid)%val,-1,Nr+1)
      call FTR_ALLOC_1_ZERO(Romp_scoeftheta(tid)%val,-1,Ntheta+1)
      call FTR_ALLOC_1_ZERO(Romp_scoefphi(tid)%val,-1,Nphi+1)
      call FTR_ALLOC_1_ZERO(Romp_scoefvpar(tid)%val,-1,Nvpar+1)
      if (conservative_scheme) then
        max_dim = MAX(Nr,Ntheta,Nphi,Nvpar)
        call FTR_ALLOC_1_ZERO(Romp_alpha(tid)%val,0,max_dim+1)
        call FTR_ALLOC_1_ZERO(Romp_rhs(tid)%val,-1,max_dim+2)
        call FTR_ALLOC_1_ZERO(Romp_vect(tid)%val,-10,10)
        call FTR_ALLOC_1_ZERO(Romp1_0Nrp1(tid)%val,0,Nr+1)
        call FTR_ALLOC_1_ZERO(Romp2_0Nrp1(tid)%val,0,Nr+1)
        call FTR_ALLOC_1_ZERO(Romp1_0Nthetap1(tid)%val,0,Ntheta+1)
        call FTR_ALLOC_2_ZERO(Romp2_0Nrp1_0Ntheta(tid)%val,0,Nr+1,0,Ntheta)
      end if
    end do

    !-> complex 1D arrays
    do tid = 1,Nbthread
      call FTR_ALLOC_1_ZERO(Comp_0Nr(tid)%val,0,Nr)
    end do

    !-> real 2D arrays (Nr,Ntheta)
    do tid = 1,Nbthread
      if (hffilterfreq .ne. 0) then
        call FTR_ALLOC_2_ZERO(Romp1_1Nrpb_1Nthetapb(tid)%val,1,2*(Nr+1),1,Ntheta+1)
      end if
      call FTR_ALLOC_2_ZERO(Romp1_1nbmom_0Nr(tid)%val,1,nbmoments,0,Nr)
    end do

    !-> complex 2D arrays (Nr,Ntheta)
    do tid = 1,Nbthread
      if (hffilterfreq .ne. 0) then
        call FTR_ALLOC_2_ZERO(Comp1_1Nrpb_1Nthetapb(tid)%val,1,2*(Nr+1),1,Ntheta)
        call FTR_ALLOC_2_ZERO(Comp1_1Nthetapb_1Nrpb(tid)%val,1,Ntheta,1,2*(Nr+1))
      end if
    end do
    call Write_dyn_balise_close('Init_OMP') !R4
!R3 call r3_info_end(r3_info_index_1) !R3
  end subroutine init_OMP


  !-----------------------------------------------------
  ! Destructor
  !-----------------------------------------------------
  subroutine del_OMP
    use globals, only : Nmu, Nbthread, hffilterfreq
    use globals, only : conservative_scheme    

    integer :: tid

    !*** array deallocations ***
    !-> 1D arrays
    do tid = 1,Nbthread
      !--> real 1D arrays
      !---> radial direction
      call FTR_DEALLOC(Romp1_0Nr(tid)%val)
      call FTR_DEALLOC(Romp2_0Nr(tid)%val)
      call FTR_DEALLOC(Romp3_0Nr(tid)%val)
      call FTR_DEALLOC(Romp4_0Nr(tid)%val)
      call FTR_DEALLOC(Romp5_0Nr(tid)%val)
      call FTR_DEALLOC(Romp6_0Nr(tid)%val)
      call FTR_DEALLOC(Romp7_0Nr(tid)%val)
      call FTR_DEALLOC(Romp8_0Nr(tid)%val)
      call FTR_DEALLOC(Romp1_1Nrp1(tid)%val)
      call FTR_DEALLOC(Romp2_1Nrp1(tid)%val)
      call FTR_DEALLOC(Romp3_1Nrp1(tid)%val)
      call FTR_DEALLOC(Romp4_1Nrp1(tid)%val)
      !---> theta direction
      call FTR_DEALLOC(Romp1_0Ntheta(tid)%val)
      call FTR_DEALLOC(Romp2_0Ntheta(tid)%val)
      call FTR_DEALLOC(Romp3_0Ntheta(tid)%val)
      call FTR_DEALLOC(Romp4_0Ntheta(tid)%val)
      call FTR_DEALLOC(Romp5_0Ntheta(tid)%val)
      call FTR_DEALLOC(Romp6_0Ntheta(tid)%val)
      call FTR_DEALLOC(Romp1_1Nthetap1(tid)%val)
      call FTR_DEALLOC(Romp2_1Nthetap1(tid)%val)
      !---> phi direction
      call FTR_DEALLOC(Romp1_0Nphi(tid)%val)
      call FTR_DEALLOC(Romp2_0Nphi(tid)%val)
      call FTR_DEALLOC(Romp3_0Nphi(tid)%val)
      call FTR_DEALLOC(Romp4_0Nphi(tid)%val)
      call FTR_DEALLOC(Romp5_0Nphi(tid)%val)
      call FTR_DEALLOC(Romp6_0Nphi(tid)%val)
      !---> vparallel direction
      call FTR_DEALLOC(Romp1_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp2_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp3_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp4_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp5_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp6_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp7_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp8_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp9_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp10_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp11_0Nvpar(tid)%val)
      call FTR_DEALLOC(Romp12_0Nvpar(tid)%val)
      !-> used for 2D advections
      call FTR_DEALLOC(Romp_vect_mpstencil(tid)%val)
      call FTR_DEALLOC(Romp_fillde_112(tid)%val)
      !--> used for cubic splines
      call FTR_DEALLOC(Romp_deriv_01(tid)%val)
      call FTR_DEALLOC(Romp_scoefr(tid)%val)
      call FTR_DEALLOC(Romp_scoeftheta(tid)%val)
      call FTR_DEALLOC(Romp_scoefphi(tid)%val)
      call FTR_DEALLOC(Romp_scoefvpar(tid)%val)
      !--> used for PSM scheme
      if (conservative_scheme) then
        call FTR_DEALLOC(Romp_alpha(tid)%val)
        call FTR_DEALLOC(Romp_rhs(tid)%val)
        call FTR_DEALLOC(Romp_vect(tid)%val)
        call FTR_DEALLOC(Romp1_0Nrp1(tid)%val)
        call FTR_DEALLOC(Romp2_0Nrp1(tid)%val)
        call FTR_DEALLOC(Romp1_0Nthetap1(tid)%val)
        call FTR_DEALLOC(Romp1_0Nrp1_0Ntheta(tid)%val)
        call FTR_DEALLOC(Romp2_0Nrp1_0Ntheta(tid)%val)
      end if
      !--> complex 1D arrays
      call FTR_DEALLOC(Comp_0Nr(tid)%val)
    end do

    !-> 2D arrays (Nr,Ntheta)
    !--> real 2D arrays (Nr,Ntheta)
    do tid = 1,Nbthread
      if (hffilterfreq .ne. 0) then
        call FTR_DEALLOC(Romp1_1Nrpb_1Nthetapb(tid)%val)
      end if
      call FTR_DEALLOC(Romp1_1nbmom_0Nr(tid)%val)
    end do

    !-> complex 2D arrays (Nr,Ntheta)
    do tid = 1,Nbthread
      if (hffilterfreq .ne. 0) then
        call FTR_DEALLOC(Comp1_1Nrpb_1Nthetapb(tid)%val)
        call FTR_DEALLOC(Comp1_1Nthetapb_1Nrpb(tid)%val)
      end if
    end do

    !*** structure deallocations ***
    !-> deallocation of structures for real 1D arrays
    !--> radial direction
    deallocate(Romp1_0Nr)
    deallocate(Romp2_0Nr)
    deallocate(Romp3_0Nr)
    deallocate(Romp4_0Nr)
    deallocate(Romp5_0Nr)
    deallocate(Romp6_0Nr)
    deallocate(Romp1_1Nrp1)
    deallocate(Romp2_1Nrp1)
    deallocate(Romp3_1Nrp1)
    deallocate(Romp4_1Nrp1)
    !-> theta direction
    deallocate(Romp1_0Ntheta)
    deallocate(Romp2_0Ntheta)
    deallocate(Romp3_0Ntheta)
    deallocate(Romp4_0Ntheta)
    deallocate(Romp5_0Ntheta)
    deallocate(Romp6_0Ntheta)
    deallocate(Romp1_1Nthetap1)
    deallocate(Romp2_1Nthetap1)
    !-> phi direction
    deallocate(Romp1_0Nphi)
    deallocate(Romp2_0Nphi)
    deallocate(Romp3_0Nphi)
    deallocate(Romp4_0Nphi)
    deallocate(Romp5_0Nphi)
    deallocate(Romp6_0Nphi)
    !-> vparallel direction
    deallocate(Romp1_0Nvpar)
    deallocate(Romp2_0Nvpar)
    deallocate(Romp3_0Nvpar)
    deallocate(Romp4_0Nvpar)
    deallocate(Romp5_0Nvpar)
    deallocate(Romp6_0Nvpar)
    deallocate(Romp7_0Nvpar)
    deallocate(Romp8_0Nvpar)
    deallocate(Romp9_0Nvpar)
    deallocate(Romp10_0Nvpar)
    deallocate(Romp11_0Nvpar)
    deallocate(Romp12_0Nvpar)
    !-> used for 2D advections
    deallocate(Romp_vect_mpstencil)
    deallocate(Romp_fillde_112)
    !--> used for cubic splines
    deallocate(Romp_deriv_01)
    deallocate(Romp_scoefr)
    deallocate(Romp_scoeftheta)
    deallocate(Romp_scoefphi)
    deallocate(Romp_scoefvpar)
    !--> used for PSM scheme
    if (conservative_scheme) then
      deallocate(Romp_alpha)
      deallocate(Romp_rhs)
      deallocate(Romp_vect)
      deallocate(Romp1_0Nrp1)
      deallocate(Romp2_0Nrp1)
      deallocate(Romp1_0Nthetap1)
      deallocate(Romp1_0Nrp1_0Ntheta)
      deallocate(Romp2_0Nrp1_0Ntheta)
    end if
    !-> deallocation of structures for complex 1D arrays
    deallocate(Comp_0Nr)
    !-> deallocation of structures for real 2D arrays (Nr,Ntheta)
    deallocate(Romp1_1Nrp1_1Nthetap1)
    deallocate(Romp2_1Nrp1_1Nthetap1)
    deallocate(Romp3_1Nrp1_1Nthetap1)
    deallocate(Romp4_1Nrp1_1Nthetap1)
    deallocate(Romp5_1Nrp1_1Nthetap1)
    deallocate(Romp6_1Nrp1_1Nthetap1)

    deallocate(Comp1_1Nrp1_1Ntheta)
    deallocate(Comp2_1Nrp1_1Ntheta)

    deallocate(Romp9_1Nrp1_1Nthetap1)
    deallocate(Romp10_1Nrp1_1Nthetap1)
    deallocate(Romp11_1Nrp1_1Nthetap1)
    deallocate(Romp12_1Nrp1_1Nthetap1)
    deallocate(Romp13_1Nrp1_1Nthetap1)
    deallocate(Romp14_1Nrp1_1Nthetap1)

    deallocate(Romp1_0Nr_0Ntheta)
    deallocate(Romp2_0Nr_0Ntheta)

    if (hffilterfreq .ne. 0) then
      deallocate(Romp1_1Nrpb_1Nthetapb)
    end if
    deallocate(Romp1_1nbmom_0Nr)

    !-> deallocation of structures for complex 2D arrays (Nr,Ntheta)
    deallocate(Comp1_1Ntheta_1Nrp1)
    if (hffilterfreq .ne. 0) then
      deallocate(Comp1_1Nrpb_1Nthetapb)
      deallocate(Comp1_1Nthetapb_1Nrpb)
    end if
  end subroutine del_OMP
end module OMPutils_module
