!*********************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!*********************************************************************

#include "config-F.inc"

!---------------------------------------------------------------------
!> \file
!> A file containing all the elements related to IO units.
!---------------------------------------------------------------------

module Io_Unit_module
  
  implicit none
  
  private
  public:: IU_open, IU_close
  
  integer, save:: lower_unit = 1
  
contains
  
  subroutine IU_open(unit, file, status, position, form, iostat)
    
    integer, intent(out):: unit
    character(len=*), intent(in):: file
    character(len=*), intent(in), optional:: status
    character(len=*), intent(in), optional:: position
    character(len=*), intent(in), optional:: form
    integer, intent(out), optional:: iostat
    
    logical:: unit_valid, unit_opened
    
    character(len=10):: r_status, r_position, r_form
    integer:: r_iostat
    
    if ( present(status) ) then
      r_status = status
    else
      r_status = 'UNKNOWN'
    endif
    if ( present(position) ) then
      r_position = position
    else
      r_position = 'ASIS'
    endif
    if ( present(form) ) then
      r_form = form
    else
      r_form = 'FORMATTED'
    endif
    
#ifdef HAS_NEWUNIT
    open(newunit=unit, file=file, status=r_status, &
            position=r_position, form=r_form, iostat=r_iostat)
#else
    unit_valid = .false.
    unit = lower_unit
    do while (.not. unit_valid)
      unit_valid = .true.
      ! Don't use units assigned to std(in|out|err)
      if ( unit == 6 .or. unit == 7 .or. unit == 10 .or. unit == 11 ) &
              then
        unit_valid = .false.
      endif
      ! Don't use already opened units
      if ( unit_valid ) then
        inquire(unit=unit, opened=unit_opened)
        if ( unit_opened ) unit_valid = .false.
      endif
      if ( .not. unit_valid ) unit = unit + 1
    enddo
    open(unit=unit, file=file, status=r_status, position=r_position, &
            form=r_form, iostat=r_iostat)
    if ( unit >= lower_unit ) lower_unit = unit + 1
#endif
    
    if ( present(iostat) ) then
      iostat = r_iostat
    endif
    
  endsubroutine IU_open
  
  subroutine IU_close(unit)
    
    integer, intent(in):: unit
    
    close(unit)
    
    if ( unit <= lower_unit ) then
      lower_unit = unit
    endif
    
  endsubroutine IU_close
  
endmodule
