!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

#include "version.h"

module version_module
  
  implicit none
  
  private
  
  public:: write_version
  
  contains
  
  !------------------------------------------------------- 
  !> Write VERSION info
  !------------------------------------------------------- 
  subroutine write_version()
    
    use globals, only : pglobal_id, outputproc, uout_res, &
                    file_name_res, code_version
    use Io_Unit_module, only: IU_open, IU_close
    
    if (pglobal_id==outputproc) then
      call IU_open(unit=uout_res,file=file_name_res,status='UNKNOWN',&
                      position='APPEND', form='FORMATTED')
      write(uout_res,*) ' '
      write(uout_res,*) '---------------------------------------------------------'
#ifdef CMAKE_VERSION
      write(uout_res,*) RCS_VERSION
      write(uout_res,'(2x,a,": ",1f6.2)') &
                      'GYSELA code version  ', code_version
      write(uout_res,'(2x,a,": ",a)') &
                      'Configured at        ', CONFIG_DATE
      write(uout_res,'(2x,a,": ",a)') &
                      'Configured by        ', CONFIG_USER
      write(uout_res,'(2x,a,": ",a)') &
                      'Configured on        ', CONFIG_MACHINE
      write(uout_res,'(2x,a,": ",a)') &
                      'Compiled at          ', COMPILE_DATE
      write(uout_res,'(2x,a,": ",a)') &
                      'Mpi Fortran Wrapper  ', MPI_FORTRAN_COMPILER
      write(uout_res,'(2x,a,": ",a)') &
                      'Mpi C Wrapper        ', MPI_C_COMPILER
      write(uout_res,'(2x,a,": ",a)') &
                      'HDF5 Fortran Wrapper ', HDF5_FORTRAN_COMPILER
      write(uout_res,'(2x,a,": ",a)') &
                      'Fortran Flags        ', FORTRAN_FLAGS
      write(uout_res,'(2x,a,": ",a)') &
                      'C Flags              ', C_FLAGS
      write(uout_res,'(2x,a,": ",a)') &
                      'Link Flags           ', LINK_FLAGS
      if (USE_WRAPPER/=0) then
        write(uout_res,*) 'Build using WRAPPER compiler'
      else
        write(uout_res,*) 'Build using compiler options for libraries'
      endif
      if (SCALASCA/=0) then
        write(uout_res,*) 'With SCALASCA instrumentation'
      else
        write(uout_res,*) 'Without instrumentation'
      endif
#else
      write(uout_res,*) &
                      ' SVN revision        : ', SVN_VERSION
      write(uout_res,'(2x,a,": ",1f6.2)') &
                      'GYSELA code version ', code_version
      write(uout_res,'(2x,a,": ",a)') &
                      'compile_time        ', trim(adjustl(compile_time))
      write(uout_res,'(2x,a,": ",a)') &
                      'compile_user        ', trim(adjustl(compile_user))
      write(uout_res,'(2x,a,": ",a)') &
                      'compile_machine     ', trim(adjustl(compile_machine))
      write(uout_res,'(2x,a,": ",a)') &
                      'fortran_compiler    ', trim(adjustl(f_compiler))
      write(uout_res,'(2x,a,": ",a)') &
                      'f_compile_flags     ', trim(adjustl(f_compile_flags))
      write(uout_res,'(2x,a,": ",a)') &
                      'c_compiler          ', trim(adjustl(c_compiler))
      write(uout_res,'(2x,a,": ",a)') &
                      'c_compile_flags     ', trim(adjustl(c_compile_flags))
      write(uout_res,'(2x,a,": ",a)') &
                      'compile_libs        ', trim(adjustl(compile_libs))
      write(uout_res,'(2x,a,": ",a)') &
                      'HDF5_version        ', trim(adjustl(HDF5_version))
#endif
      write(uout_res,*) '---------------------------------------------------------'
      write(uout_res,*) ' '
    end if
    
    call IU_close(unit=uout_res)
    
  end subroutine write_version

endmodule version_module
